/*
 * Copyright (c) 1999 Robert N. Watson
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *       $Id: $
 */

/*
 * Support for POSIX.1e.AUD auditing support
 */

#include "opt_audit.h"

#include <sys/param.h>
#include <sys/types.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/proc.h>
#include <sys/vnode.h>
#include <sys/filio.h>
#include <sys/ttycom.h>
#include <sys/signalvar.h>
#include <sys/syslog.h>
#include <sys/sysctl.h>
#include <sys/syscall.h>
#include <sys/sysproto.h>
#include <sys/sysent.h>
#include <sys/kernel.h>
#include <sys/poll.h>
#include <sys/filedesc.h>
#include <sys/malloc.h>
#include <sys/audit.h>

struct audit_func af_array;

#ifdef _POSIX_AUD

/* #include <security/audit/audit_private.h> */

#define AUDIT_RDPRI	(PZERO + 1)

#define AUDIT_ASYNC	0x04
#define AUDIT_RDWAIT	0x08

#define	MIN(a,b)	(((a)<(b))?(a):(b))

#define AUDIT_MAX_AUD_WRITE_LEN	4096

#define AUDIT_MAJOR	120
#define AUDIT_MINOR	0

SYSCTL_DECL(_security);

SYSCTL_NODE(_security, OID_AUTO, audit, CTLFLAG_RD, 0,
    "TrustedBSD Audit Subtree");

static int audit_open = 0;
SYSCTL_UINT(_security_audit, OID_AUTO, enabled, CTLFLAG_RD,
    &audit_open, 0, "");

/*
 * Prototypes
 */

static d_open_t		audit_device_open;
static d_close_t	audit_device_close;
static d_read_t		audit_device_read;
static d_ioctl_t	audit_device_ioctl;
static d_poll_t		audit_device_poll;

static void audit_wakeup(void);
static void audit_flush(void);
/*
 * Support for /dev/audit
 */

static struct cdevsw audit_device_cdevsw = { 
	.d_open =	audit_device_open,	
	.d_close =	audit_device_close,
	.d_read =	audit_device_read,
	.d_ioctl =	audit_device_ioctl,
	.d_poll =	audit_device_poll,
	.d_name =	"audit",
	.d_maj =	AUDIT_MAJOR,
};

static struct auditsoftc {
	int	sc_state;		/* see above for possibilities */
	struct	selinfo sc_selp;	/* process waiting on select call */
	struct	sigio *sc_sigio;	/* information for async I/O */
} auditsoftc;

static dev_t audit_dev;

/* Lock mutex. This mutex locks audit buffer  audit_list */
struct mtx audit_mtx;

/*
 * List to hold buffered records, stored in a form appropriate for immediate
 * copy into userland.
 */
static TAILQ_HEAD(audit_list, audit_list_entry) audit_list_head;
struct audit_list_entry {
	char	*aud_rec_data;		/* data blob */
	int	aud_rec_data_len;	/* length of data blob */

	struct flat_header	fh;	/* the header, seperable */
	int	fh_read_header;		/* the header has been read */

	TAILQ_ENTRY(audit_list_entry)	audit_list_entries;
};

/*
 * Structure to hold a record while under construction in the kernel
 */
struct audit_record {
	uid_t	aud_rec_uid;
	char	*aud_hdr_data, *aud_evinfo_data, *aud_obj_data, *aud_subj_data;
	int	aud_hdr_data_len, aud_evinfo_data_len, aud_obj_data_len,
		aud_subj_data_len;
	int	aud_hdr_data_buflen, aud_evinfo_data_buflen,
		aud_obj_data_buflen, aud_subj_data_buflen;
	int	status_set;
};

MALLOC_DEFINE(M_AUDBUF, "auditbuf",
	"Buffers for userland and kernel audit data");
MALLOC_DEFINE(M_AUDMISC,"auditdata","Private audit structures");

static int
audit_modevent(module_t mod, int type, void *data)
{
	switch(type) {
	case MOD_LOAD:
		printf("TrustedBSD AUDIT Subsystem loading ");
		audit_dev = make_dev(&audit_device_cdevsw, AUDIT_MINOR,
				UID_ROOT, GID_WHEEL, 0666, "audit");
		audit_open = 0;
		mtx_init(&audit_mtx, "audit_list lock", NULL, MTX_DEF);
		TAILQ_INIT(&audit_list_head);
		bzero(&af_array, sizeof(struct audit_func));
		printf("success.\n");
		return 0;

	case MOD_UNLOAD:
		audit_flush();
		audit_open = 0;
		destroy_dev(audit_dev);
		mtx_destroy(&audit_mtx);
		printf("TrustedBSD AUDIT Subsystem unloading success.\n");
		return 0;

	case MOD_SHUTDOWN:
		
		return 0;

	default:
		return EOPNOTSUPP;
	}
}


/*
 * Create a new in-kernel audit event, holding uid responsible, and of type
 * event_type (see sys/audit.h).  Uses M_ZERO, so should not fail?
 */
void *
new_audit_event(int event_type, uid_t uid, aud_id_t audit_id)
{
	struct flat_aud_hdr	*f_hdr;
	struct flat_aud_info	*f_info;
	struct audit_record	*rec;
	aud_id_t	*anaud_id_t;
	aud_time_t	*anaud_time_t;
	struct timespec	ts;
	char	*current; /* changed it from (void *) */
	short	*ashort;
	int	*anint;

	rec = malloc(sizeof(struct audit_record), M_AUDBUF, M_ZERO);
	if (rec == NULL)
		return(NULL);

	rec->aud_evinfo_data = rec->aud_obj_data = rec->aud_subj_data = NULL;

	rec->aud_evinfo_data_len = rec->aud_obj_data_len =
	    rec->aud_subj_data_len = 0;
	rec->aud_evinfo_data_buflen = rec->aud_obj_data_buflen =
	    rec->aud_subj_data_buflen = 0;

	/* leave room for status, errno to be filled in later */
	rec->aud_hdr_data_len =
	    sizeof(struct flat_aud_hdr) +
	    sizeof(struct flat_aud_info) +
	    sizeof(short) +			/* AUD_FORMAT */
	    sizeof(struct flat_aud_info) +
	    sizeof(short) +			/* AUD_VERSION */
	    sizeof(struct flat_aud_info) +
	    sizeof(aud_id_t) +			/* AUD_AUD_ID */
	    sizeof(struct flat_aud_info) +
	    sizeof(int) +			/* AUD_EVENT_TYPE */
	    sizeof(struct flat_aud_info) +
	    sizeof(aud_time_t);		 	/* AUD_TIME */

	/* here's the status + errno */
	rec->aud_hdr_data_buflen =
	    rec->aud_hdr_data_len +
	    sizeof(struct flat_aud_info) +
	    sizeof(aud_status_t) +		/* AUD_STATUS */
	    sizeof(struct flat_aud_info) +
	    sizeof(int);			/* AUD_ERRNO */

	rec->aud_hdr_data = malloc(rec->aud_hdr_data_buflen, M_AUDBUF,
	    M_ZERO);
	if (rec->aud_hdr_data == NULL) {
		free(rec, M_AUDBUF);
		return(NULL);
	}

	/*
	 * HDR
	 */
	current = rec->aud_hdr_data;
	f_hdr = (struct flat_aud_hdr *)current;
	f_hdr->fah_magic = AUD_MAGIC_FHDR;
	f_hdr->fah_num_info = -1;
	current += sizeof(struct flat_aud_hdr);

	/*
	 * AUD_FORMAT
	 */
	f_info = (struct flat_aud_info *)current;
	f_info->fai_magic = AUD_MAGIC_FINFO;
	f_info->fai_type = AUD_TYPE_SHORT;
	f_info->fai_item_id = AUD_FORMAT;
	f_info->fai_length = sizeof(short);
	current += sizeof(struct flat_aud_info);

	ashort = (short *)current;
	*ashort = AUD_NATIVE;
	current += sizeof(short);

	/*
	 * AUD_VERSION
	 */
	f_info = (struct flat_aud_info *)current;
	f_info->fai_magic = AUD_MAGIC_FINFO;
	f_info->fai_type = AUD_TYPE_SHORT;
	f_info->fai_item_id = AUD_VERSION;
	f_info->fai_length = sizeof(short);
	current += sizeof(struct flat_aud_info);

	ashort = (short *)current;
	*ashort = AUD_STD_NOT;
	current += sizeof(short);

	/*
	 * AUD_AUD_ID
	 */
	f_info = (struct flat_aud_info *)current;
	f_info->fai_magic = AUD_MAGIC_FINFO;
	f_info->fai_type = AUD_TYPE_AUD_ID;
	f_info->fai_item_id = AUD_AUD_ID;
	f_info->fai_length = sizeof(aud_id_t);
	current += sizeof(struct flat_aud_info);

	anaud_id_t = (aud_id_t *)current;
	*anaud_id_t = audit_id;
	current += sizeof(aud_id_t);

	/*
	 * AUD_EVENT_TYPE
	 */
	f_info = (struct flat_aud_info *)current;
	f_info->fai_magic = AUD_MAGIC_FINFO;
	f_info->fai_type = AUD_TYPE_INT;
	f_info->fai_item_id = AUD_EVENT_TYPE;
	f_info->fai_length = sizeof(int);
	current += sizeof(struct flat_aud_info);

	anint = (int *)current;
	*anint = event_type;
	current += sizeof(int);

	/*
	 * AUD_TIME
	 */
	f_info = (struct flat_aud_info *)current;
	f_info->fai_magic = AUD_MAGIC_FINFO;
	f_info->fai_type = AUD_TYPE_AUD_TIME;
	f_info->fai_item_id = AUD_TIME;
	f_info->fai_length = sizeof(aud_time_t);
	current += sizeof(struct flat_aud_info);

	anaud_time_t = (aud_time_t *)current;
	getnanotime(&ts);
	anaud_time_t->sec = ts.tv_sec;
	anaud_time_t->nsec = ts.tv_nsec;
	current += sizeof(aud_time_t);

	rec->status_set = 0;
	rec->aud_rec_uid = uid;

	if (current > rec->aud_hdr_data + rec->aud_hdr_data_len)
		panic("new_aud_event.invariant_failed\n");

	return(rec);
}


/*
 * Put a piece of information into event, section {hdr,evinfo,obj,subj} of
 * type, located at dataptr
 */
int
audit_put(void *event, int section, int item_id, int type, int len,
	  void *dataptr)
{
	struct audit_record	*rec;
	struct flat_aud_info	*f_info;
	struct flat_aud_hdr	*f_hdr;
	struct flat_aud_evinfo	*f_evinfo;
	struct flat_aud_obj	*f_obj;
	struct flat_aud_subj	*f_subj;
	void	*newchunk;
	char	*current;
	int	new_len;

	if (!event) {
		printf("audit_put: null event\n");
		return(EINVAL);
	}
	rec = event;

	switch(section) {
	case AUD_MAGIC_HDR:
		new_len = len + sizeof(struct flat_aud_info) +
		    rec->aud_hdr_data_len;
		if (rec->aud_hdr_data_len == 0) {
			/* need space for flat_aud_hdr */
			new_len += sizeof(struct flat_aud_hdr);
		}
		if (new_len > rec->aud_hdr_data_buflen) {
			/* will need more space */
			newchunk = malloc(new_len, M_AUDBUF, M_ZERO);
			if (newchunk == NULL)
				return(ENOMEM);
			if (rec->aud_hdr_data != NULL) {
				bcopy(rec->aud_hdr_data, newchunk,
				    rec->aud_hdr_data_len);
				free(rec->aud_hdr_data, M_AUDBUF);
			}
			rec->aud_hdr_data = newchunk;
			rec->aud_hdr_data_buflen = new_len;
		}
		if (rec->aud_hdr_data_len == 0) {
			f_hdr = (struct flat_aud_hdr *)rec->aud_hdr_data;
			f_hdr->fah_magic = AUD_MAGIC_HDR;
			f_hdr->fah_num_info = -1;
			rec->aud_hdr_data_len +=
			    sizeof(struct flat_aud_hdr);
		}
		current = rec->aud_hdr_data + rec->aud_hdr_data_len;
		break;

	case AUD_MAGIC_EVINFO:
		new_len = len + sizeof(struct flat_aud_info) +
		    rec->aud_evinfo_data_len;
		if (rec->aud_evinfo_data_len == 0) {
			/* will need space for a flat_aud_evinfo */
			new_len += sizeof(struct flat_aud_evinfo);
		}
		if (new_len > rec->aud_evinfo_data_buflen) {
			/* will need more space */
			newchunk = malloc(new_len, M_AUDBUF, M_ZERO);
			if (newchunk == NULL)
				return(ENOMEM);
			if (rec->aud_evinfo_data != NULL) {
				bcopy(rec->aud_evinfo_data, newchunk,
				    rec->aud_evinfo_data_len);
				free(rec->aud_evinfo_data, M_AUDBUF);
			}
			rec->aud_evinfo_data = newchunk;
			rec->aud_evinfo_data_buflen = new_len;
		}
		if (rec->aud_evinfo_data_len == 0) {
			f_evinfo = (struct flat_aud_evinfo *)rec->aud_evinfo_data;
			f_evinfo->fae_magic = AUD_MAGIC_FEVINFO;
			f_evinfo->fae_num_info = -1;
			rec->aud_evinfo_data_len +=
			    sizeof(struct flat_aud_evinfo);
		}
		current = rec->aud_evinfo_data + rec->aud_evinfo_data_len;
		break;

	case AUD_MAGIC_OBJ:
		new_len = len + sizeof(struct flat_aud_info) +
		    rec->aud_obj_data_len;
		if (rec->aud_obj_data_len == 0) {
			/* will need space for a flat_aud_obj */
			new_len += sizeof(struct flat_aud_obj);
		}
		if (new_len > rec->aud_obj_data_buflen) {
			/* will need more space */
			newchunk = malloc(new_len, M_AUDBUF, M_ZERO);
			if (newchunk == NULL)
				return(ENOMEM);
			if (rec->aud_obj_data != NULL) {
				bcopy(rec->aud_obj_data, newchunk,
				    rec->aud_obj_data_len);
				free(rec->aud_obj_data, M_AUDBUF);
			}
			rec->aud_obj_data = newchunk;
			rec->aud_obj_data_buflen = new_len;
		}
		if (rec->aud_obj_data_len == 0) {
			f_obj = (struct flat_aud_obj *)rec->aud_obj_data;
			f_obj->fao_magic = AUD_MAGIC_FOBJ;
			f_obj->fao_num_info = -1;
			rec->aud_obj_data_len +=
			    sizeof(struct flat_aud_obj);
		}
		current = rec->aud_obj_data + rec->aud_obj_data_len;
		break;

	case AUD_MAGIC_SUBJ:
		new_len = len + sizeof(struct flat_aud_info) +
		    rec->aud_subj_data_len;
		if (rec->aud_subj_data_len == 0) {
			/* will need space for a flat_aud_subj */
			new_len += sizeof(struct flat_aud_subj);
		}
		if (new_len > rec->aud_subj_data_buflen) {
			/* will need more space */
			newchunk = malloc(new_len, M_AUDBUF, M_ZERO);
			if (newchunk == NULL)
				return(ENOMEM);
			if (rec->aud_subj_data) {
				bcopy(rec->aud_subj_data, newchunk,
				    rec->aud_subj_data_len);
				free(rec->aud_subj_data, M_AUDBUF);
			}
			rec->aud_subj_data = newchunk;
			rec->aud_subj_data_buflen = new_len;
		}
		if (rec->aud_subj_data_len == 0) {
			f_subj = (struct flat_aud_subj *)rec->aud_subj_data;
			f_subj->fas_magic = AUD_MAGIC_FSUBJ;
			f_subj->fas_num_info = -1;
			rec->aud_subj_data_len +=
			    sizeof(struct flat_aud_subj);
		}
		current = rec->aud_subj_data + rec->aud_subj_data_len;
		break;

	default:
		return(EINVAL);
	}

	/* there is enough room for the flat_aud_info and the data itself */
	f_info = (struct flat_aud_info *)current;
	f_info->fai_magic = AUD_MAGIC_FINFO;
	f_info->fai_type = type;
	f_info->fai_item_id = item_id;

	switch(type) {
	case AUD_TYPE_STRING_ARRAY:
		printf("audit_put: don't know how to deal with arrays yet\n");
		return(EINVAL);
	default:
		f_info->fai_length = len;
		current += sizeof(struct flat_aud_info);
		bcopy(dataptr, current, len);
		current += len;
	}

	switch(section) {
	case AUD_MAGIC_HDR:
		rec->aud_hdr_data_len += sizeof(struct flat_aud_info) + len;
		break;
	case AUD_MAGIC_EVINFO:
		rec->aud_evinfo_data_len += sizeof(struct flat_aud_info) + len;
		break;
	case AUD_MAGIC_OBJ:
		rec->aud_obj_data_len += sizeof(struct flat_aud_info) + len;
		break;
	case AUD_MAGIC_SUBJ:
		rec->aud_subj_data_len += sizeof(struct flat_aud_info) + len;
		break;
	default:
		/* shouldn't get here, as just did this above! */
		panic("unknown section");
	}

	return(0);
}

/* Type         | Description                   | item_id
 * -------------+-------------------------------+-----------------------------
 * AUD_TYPE_PID | The process ID                | AUD_PID
 * AUD_TYPE_UID | The effective user ID         | AUD_EUID
 * AUD_TYPE_GID | The effective group ID        | AUD_EGID
 * AUD_TYPE_GID | The supplementary group IDs   | AUD_SGIDS
 * AUD_TYPE_UID | The real user ID              | AUD_RUID
 * AUD_TYPE_GID | The real group ID             | AUD_RGID
 * AUD_TYPE_MAC | The process MAC label         | AUD_MAC_LBL
 * AUD_TYPE_INF | The process information label | AUD_INF_LBL
 * AUD_TYPE_CAP | The process capability state  | AUD_CAP
 */

int
audit_put_subj(void *event, struct thread *td)
{
	int aud_cnt = 0;
	/* XXX: Need to understand record format, userland doesn't
	 * understand it right now!
	 */
#if 0
	aud_cnt += audit_put(event, AUD_MAGIC_SUBJ, AUD_PID, AUD_TYPE_PID,
			 sizeof(pid_t), &td->td_proc->p_pid);
	aud_cnt += audit_put(event, AUD_MAGIC_SUBJ, AUD_EUID, AUD_TYPE_UID,
			 sizeof(uid_t), &td->td_ucred->cr_uid);
	aud_cnt += audit_put(event, AUD_MAGIC_SUBJ, AUD_EGID, AUD_TYPE_GID,
			 sizeof(gid_t), &td->td_ucred->cr_gid);

	aud_cnt += audit_put(event, AUD_MAGIC_SUBJ, AUD_SGIDS, AUD_TYPE_GID,
			 NGROUPS*sizeof(gid_t), &td->td_ucred->cr_groups);

	aud_cnt += audit_put(event, AUD_MAGIC_SUBJ, AUD_RUID, AUD_TYPE_UID,
			 sizeof(uid_t), &td->td_ucred->cr_ruid);
	aud_cnt += audit_put(event, AUD_MAGIC_SUBJ, AUD_RGID, AUD_TYPE_GID,
			 sizeof(gid_t), &td->td_ucred->cr_rgid);

	aud_cnt += audit_put(event, AUD_MAGIC_SUBJ, AUD_MAC_LBL, AUD_TYPE_MAC,
			 sizeof(mac_t), &!!!));

/*audit_put(void *event, int section,
		 int item_id,
		 int type,
		 int len,
		 void *dataptr)*/
#endif
	return (aud_cnt);
}

int
audit_put_obj(void *event, struct thread *td)
{

	return (0);
}

/*
 * Submit a completed audit record for delivery to userland.
 */
int
audit_submit(void *event)
{
	struct audit_list_entry	*ale;
	struct audit_record	*rec;
	char	*current;
	int	error;

	rec = event;

	/* don't create new records while the device
	 * is not initialized or openned
	*/
	if(!mtx_initialized(&audit_mtx) && !audit_open) {
		error = ENXIO; /* XXX: ??? */
		goto error;
	}

	ale = malloc(sizeof(struct audit_list_entry), M_AUDBUF, M_ZERO);
	if (ale == NULL) {
		error = ENOMEM;
		goto error;
	}

	ale->aud_rec_data_len = rec->aud_hdr_data_len +
	    rec->aud_evinfo_data_len + rec->aud_obj_data_len +
	    rec->aud_subj_data_len;

	if (!ale->aud_rec_data_len) {
		free(ale, M_AUDBUF);
		error = EINVAL;
		goto error;
	}

	ale->aud_rec_data = malloc(ale->aud_rec_data_len, M_AUDBUF, M_ZERO);
	if (ale->aud_rec_data == NULL) {
		free(ale, M_AUDBUF);
		error = ENOMEM;
		goto error;
	}

	/* Flattened Record Header */
	ale->fh.fh_magic = AUD_MAGIC_FLAT;
	ale->fh.fh_length = ale->aud_rec_data_len;
	ale->fh.fh_version = AUD_FLAT_VERSION;
	ale->fh_read_header = 0;
	ale->fh.fh_uid = rec->aud_rec_uid;

	current = ale->aud_rec_data;

	if (rec->aud_hdr_data) {
		memcpy(current, rec->aud_hdr_data, rec->aud_hdr_data_len);
		current += rec->aud_hdr_data_len;
	}

	if (rec->aud_evinfo_data) {
		memcpy(current, rec->aud_evinfo_data, rec->aud_evinfo_data_len);
		current += rec->aud_evinfo_data_len;
	}

	if (rec->aud_obj_data) {
		memcpy(current, rec->aud_obj_data, rec->aud_obj_data_len);
		current += rec->aud_obj_data_len;
	}

	if (rec->aud_subj_data) {
		memcpy(current, rec->aud_subj_data, rec->aud_subj_data_len);
		current += rec->aud_subj_data_len;
	}

	if (current > ale->aud_rec_data + ale->aud_rec_data_len)
		panic("audit_submit.invariant failed\n");
	AUDIT_LOCK(); /* __LINE__ + 2 */
	TAILQ_INSERT_TAIL(&audit_list_head, ale, audit_list_entries);
	AUDIT_UNLOCK();
	audit_wakeup();

	error = 0;

error:
	if (rec->aud_hdr_data)
		free(rec->aud_hdr_data, M_AUDBUF);
	if (rec->aud_evinfo_data)
		free(rec->aud_evinfo_data, M_AUDBUF);
	if (rec->aud_obj_data)
		free(rec->aud_obj_data, M_AUDBUF);
	if (rec->aud_subj_data)
		free(rec->aud_subj_data, M_AUDBUF);
	free(rec, M_AUDBUF);

	return(error);
}

static void
audit_flush(void)
{
	AUDIT_LOCK(); /* __LINE__ + 4 */
	while (TAILQ_FIRST(&audit_list_head) != NULL)
		TAILQ_REMOVE(&audit_list_head, audit_list_head.tqh_first,
		    audit_list_entries);
	AUDIT_UNLOCK();
}

/*
 * AUDIT driver implementation
 */

static int
audit_device_close(dev_t dev, int flags, int fmt, struct thread *td)
{

	if (minor(dev) != 0)
	    return(ENXIO);

	if (!audit_open)
	    panic("audit_close: not open");

	audit_open = 0;
	auditsoftc.sc_state = 0;
	funsetown(&auditsoftc.sc_sigio);

	printf("Auditing log flushed and disabled\n");
	audit_flush();

	return(0);
}

static int
audit_device_open(dev_t dev, int flags, int fmt, struct thread *td)
{

	if (minor(dev) != 0)
	    return(ENXIO);

	if (audit_open) {
	    printf("audit_open: device already open\n");
	    return(EBUSY);
	}

	/* XXX: Does lock need here? */
	AUDIT_LOCK(); /* __LINE__ + 5 */
	audit_open = 1;

	/* signal process only */
	fsetown(td->td_proc->p_pid, &auditsoftc.sc_sigio);
	AUDIT_UNLOCK();

	/*
	 * reset first record in chain so that it is read for scratch, as
	 * the last listener did not pick up the whole record
	 */

	return(0);
}

static int
audit_device_read(dev_t dev, struct uio *uio, int flags)
{
	struct audit_list_entry	*ale;
	void	*data;
	int	amnt, rv, datalen;
	/*int	s;*/

	rv = 0;

	/*s = splhigh();*/
	AUDIT_LOCK(); 
	while ((ale = TAILQ_FIRST(&audit_list_head)) == NULL) { 
		AUDIT_UNLOCK();
		if (flags & IO_NDELAY) {
	 		/*splx(s);*/
	 		return(EWOULDBLOCK);
	 	}
	  	auditsoftc.sc_state |= AUDIT_RDWAIT;
	  	if ((rv = tsleep((caddr_t) &audit_list_head, AUDIT_RDPRI |
		    PCATCH, "audit", 0))) {
	 		/*splx(s);*/
			return(rv);
		}
		AUDIT_LOCK(); 
	}
	/*splx(s);*/
	AUDIT_UNLOCK();
	auditsoftc.sc_state &= ~AUDIT_RDWAIT;

	/*
	 * perform copy
	 */
	if (!ale->fh_read_header) {
		data = &(ale->fh);
		datalen = sizeof(struct flat_header);
	} else {
		data = ale->aud_rec_data;
		datalen = ale->aud_rec_data_len;
	}
	while (uio->uio_resid > 0) {
		amnt = MIN(uio->uio_resid, datalen);
		if (amnt == 0)
			break;
		if ((rv = uiomove((caddr_t)data, amnt, uio)) != 0)
			break;
		datalen -= amnt;
	}

	if (datalen != 0) {
		printf("audit_device_read: not all data read in readv()\n");
		/*
		 * junk the remainder of the record if that was just
		 * the flat_header 
	 	 */
		ale->fh_read_header = 1;
	} else {
	}

	if (!ale->fh_read_header) {
		ale->fh_read_header = 1;
	} else {
		AUDIT_LOCK(); /* __LINE__ + 3 */
		TAILQ_REMOVE(&audit_list_head, audit_list_head.tqh_first,
		    audit_list_entries);
		AUDIT_UNLOCK();
		free(ale->aud_rec_data, M_AUDBUF);
		free(ale, M_AUDBUF);
	}
	return(rv);
}

static int
audit_device_ioctl(dev_t dev, u_long cmd, caddr_t data, int flags,
	struct thread *td)
{

	return(ENOTTY);
}

static int
audit_device_poll(dev_t dev, int events, struct thread *td)
{
	/*int s;*/
	int revents = 0;

	/*s = splhigh();*/
	AUDIT_LOCK(); /* __LINE__ + 10 */

	if (events & (POLLIN | POLLRDNORM)) {
		if (!TAILQ_EMPTY(&audit_list_head)) {
			 revents |= events & (POLLIN | POLLRDNORM);
		} else {
			selrecord(td, &auditsoftc.sc_selp);
		}
	}
	/*splx(s);*/
	AUDIT_UNLOCK();
	return (revents);
}

static void
audit_wakeup(void)
{

	if (!audit_open)
		return;

	selwakeup(&auditsoftc.sc_selp);

	if ((auditsoftc.sc_state & AUDIT_ASYNC) &&
	    (auditsoftc.sc_sigio != NULL))
		pgsigio(&auditsoftc.sc_sigio, SIGIO, 0);

	if (auditsoftc.sc_state & AUDIT_RDWAIT) {
		wakeup((caddr_t) &audit_list_head);
		auditsoftc.sc_state &= ~AUDIT_RDWAIT;
	}
}

/*
 * syscall implementations
 */
int
__aud_switch_sc(td, uap)
	struct thread *td;
	struct __aud_switch_sc_args *uap;
{
	aud_state_t	aud_state_in, aud_state_out;
	int	error;

	error = copyin(uap->aud_state, &aud_state_in, sizeof(aud_state_t));
	if (error) {
		if (AUDIT_PROC(td->td_proc))
			AUDIT_AUD_SWITCH(td, AUD_FAIL_OTHER, error,
			    aud_state_in);
		return(error);
	}

	switch(aud_state_in) {
	case AUD_STATE_OFF:
		error = suser(td);
		if (error) {
			if (AUDIT_PROC(td->td_proc))
				AUDIT_AUD_SWITCH(td, AUD_FAIL_OTHER, error,
				    aud_state_in);
			return(error);
		}
		/* XXX: Need to lock? */
		PROC_LOCK(td->td_proc);
		td->td_proc->p_posix1e_flag &= ~POSIX1E_FLAG_AUDIT;
		PROC_UNLOCK(td->td_proc);
		break;
	case AUD_STATE_ON:
		error = suser(td);
		if (error) {
			if (AUDIT_PROC(td->td_proc))
				AUDIT_AUD_SWITCH(td, AUD_FAIL_OTHER, error,
				    aud_state_in);
			return(error);
		}
		/* XXX: Need to lock? */
		PROC_LOCK(td->td_proc);
		td->td_proc->p_posix1e_flag |= POSIX1E_FLAG_AUDIT;
		PROC_UNLOCK(td->td_proc);
		break;
	case AUD_STATE_QUERY:
		/* XXX: Need to lock? */
		if (td->td_proc->p_posix1e_flag & POSIX1E_FLAG_AUDIT)
			aud_state_out = AUD_STATE_ON;
		else
			aud_state_out = AUD_STATE_OFF;
		error = copyout(&aud_state_out, uap->aud_state,
		    sizeof(aud_state_t));
		if (error) {
			if (AUDIT_PROC(td->td_proc))
				AUDIT_AUD_SWITCH(td, AUD_FAIL_OTHER, error,
				    aud_state_in);
			return(error);
		}
		break;
	default:
		if (AUDIT_PROC(td->td_proc))
			AUDIT_AUD_SWITCH(td, AUD_FAIL_OTHER, EINVAL,
			    aud_state_in);
		return(EINVAL);
	}

	if (AUDIT_PROC(td->td_proc))
		AUDIT_AUD_SWITCH(td, AUD_SUCCESS, 0, aud_state_in);
	return(0);
}

int
__aud_write_sc(td, uap)
	struct thread *td;
	struct __aud_write_sc_args *uap;
{
	struct audit_list_entry	*ale;
	char	*buf;
	int	i;

	i = suser(td);
	if (i) {
		AUDIT_AUD_WRITE(td, AUD_FAIL_PRIV, i);
		return(i);
	}

	/* require to be > sizeof the header */
	if (uap->len <= sizeof(struct flat_header)) {
		AUDIT_AUD_WRITE(td, AUD_FAIL_OTHER, EINVAL);
		return(EINVAL);
	}
	if (uap->len > AUDIT_MAX_AUD_WRITE_LEN) {
		AUDIT_AUD_WRITE(td, AUD_FAIL_OTHER, ENOMEM);
		return(ENOMEM);
	}

	ale = (struct audit_list_entry *)
	    malloc(sizeof(struct audit_list_entry),  M_AUDBUF, M_ZERO);
	if (!ale) {
		AUDIT_AUD_WRITE(td, AUD_FAIL_OTHER, ENOMEM);
		return(ENOMEM);
	}

	/* always at least one byte */
	buf = (char *) malloc(uap->len - sizeof(struct flat_header), M_AUDBUF,
	    M_ZERO);
	if (!buf) {
		free(ale, M_AUDBUF);
		AUDIT_AUD_WRITE(td, AUD_FAIL_OTHER, ENOMEM);
		return(ENOMEM);
	}

	i = copyin(uap->addr, &(ale->fh), sizeof(struct flat_header));
	if (i) {
		free(ale, M_AUDBUF);
		free(buf, M_AUDBUF);
		AUDIT_AUD_WRITE(td, AUD_FAIL_OTHER, i);
		return(i);
	}

	if (ale->fh.fh_magic != AUD_MAGIC_FLAT) {
		printf("aud_write_sc: bad magic on submitted audit record\n");
		free(ale, M_AUDBUF);
		free(buf, M_AUDBUF);
		AUDIT_AUD_WRITE(td, AUD_FAIL_OTHER, EINVAL);
		return(EINVAL);
	}

	if (ale->fh.fh_length != uap->len - sizeof(struct flat_header)) {
		printf("aud_write_sc: ale->fh.fh_length=%d, uap->len=%d\n",
		    ale->fh.fh_length, uap->len);
		free(ale, M_AUDBUF);
		free(buf, M_AUDBUF);
		AUDIT_AUD_WRITE(td, AUD_FAIL_OTHER, EINVAL);
		return(EINVAL);
	}

	i = copyin(uap->addr + sizeof(struct flat_header), buf, uap->len -
	    sizeof(struct flat_header));
	if (i) {
		free(ale, M_AUDBUF);
		free(buf, M_AUDBUF);
		AUDIT_AUD_WRITE(td, AUD_FAIL_OTHER, i);
		return(i);
	}

	ale->fh.fh_uid = td->td_ucred->cr_uid;
	ale->fh_read_header = 0;
	ale->aud_rec_data_len = uap->len - sizeof(struct flat_header);
	ale->aud_rec_data = buf;

	AUDIT_LOCK(); /* __LINE__ + 2 */
	TAILQ_INSERT_TAIL(&audit_list_head, ale, audit_list_entries);
	AUDIT_UNLOCK();
	audit_wakeup();

	/* don't need to audit success */
	return(0);
}

int
__aud_get_id_sc(td, uap)
	struct thread *td;
	struct __aud_get_id_sc_args *uap;
/*
	pid_t pid;
	uid_t * uid;
*/
{
	struct proc *p = pfind(uap->pid);

	/* Need lock? */
	PROC_LOCK(p);
	copyout(&p->p_audit_id, uap->uid, sizeof(aud_id_t));
	PROC_UNLOCK(p);
	return (0);
}

int
__aud_set_id_sc(td, uap)
	struct thread *td;
	struct __aud_set_id_sc_args *uap;
/*
	uid_t uid;
*/
{
	/* Needs change to capability cap_check! */
	int error = suser(td);
	if(!error) {
		PROC_LOCK(td->td_proc);
		td->td_proc->p_audit_id = uap->uid;
		PROC_UNLOCK(td->td_proc);
	}
	AUDIT_AUD_SET_ID(td, AUD_FAIL_OTHER, error, uap->uid);
	return (error);
}

static moduledata_t audit_mod = {
	"audit",
	&audit_modevent,
	NULL
};

SYSCALL_MODULE_HELPER(__aud_switch_sc);
SYSCALL_MODULE_HELPER(__aud_write_sc);
SYSCALL_MODULE_HELPER(__aud_get_id_sc);
SYSCALL_MODULE_HELPER(__aud_set_id_sc);
/* XXX: is it needed? */

DECLARE_MODULE(audit, audit_mod, SI_SUB_DRIVERS, SI_ORDER_MIDDLE+AUDIT_MAJOR);

MODULE_VERSION(audit, 1);

#else /* !_POSIX_AUD */
/*
int
__aud_switch_sc(td, uap)
	struct thread *td;
	struct __aud_switch_sc_args *uap;
{

	return(ENOSYS);
}

int
__aud_write_sc(td, uap)
	struct thread *td;
	struct __aud_write_sc_args *uap;
{

	return(ENOSYS);
}

int
__aud_get_id_sc(td, uap)
	struct thread *td;
	struct __aud_get_id_sc_args *uap;
{

	return(ENOSYS);
}
*/
#endif /* _POSIX_AUD */
